from logging import log
from pathlib import Path
from typing import Optional, Tuple, Union, List
from beastiary import crud, schemas
from beastiary.db.database import Database
from pydantic.utils import is_valid_field
from beastiary.log import logger
import os, math, errno


def get_headers(path: Path, delimiter: Optional[str] = "\t") -> Tuple[int, str]:
    logger.debug(f"Getting headers from {path}")
    with open(path, "r") as f:
        while True:
            line = f.readline()
            if line.startswith("#") or line.startswith("["):
                continue
            headers_list = line.strip().split(delimiter)
            try:
                headers_list[0] = "state"
            except IndexError:
                raise ValueError(f"Could not find headers in {path}")
            headers_line = delimiter.join(headers_list)
            last_byte = f.tell()
            logger.debug(f"last_byte = {last_byte}")
            logger.debug(f"headers_line = {headers_line}")
            logger.debug(f"delimiter = {delimiter}")
            return last_byte, headers_line


def is_valid_log_file(headers_line: str, delimiter: Optional[str] = None) -> bool:
    if len(headers_line.split(delimiter)) > 1:
        return True
    return False


def add_trace(db: dict, trace_in: schemas.TraceCreate) -> dict:
    if not trace_in.path.is_file():
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), trace_in.path)
    last_byte, headers_line = get_headers(trace_in.path, delimiter=trace_in.delimiter)
    if not is_valid_log_file(headers_line, delimiter=trace_in.delimiter):
        raise ValueError(f"Invalid log file: {trace_in.path}")
    logger.debug(f"Creating trace: {trace_in}")
    trace = crud.trace.create(
        db, obj_in=trace_in, last_byte=last_byte, headers_line=headers_line
    )
    logger.debug(f"Created trace: {trace}")
    return trace


def read_lines(trace: schemas.Trace) -> Tuple[int, list]:
    logger.debug(f"reading lines from: {trace}")
    if not trace["path"]:
        raise ValueError("Path must be set.")
    with open(trace["path"], "r") as f:
        f.seek(trace["last_byte"], 0)
        lines = f.readlines()
        if lines:
            last_byte = f.tell()
        else:
            last_byte = trace["last_byte"]
        logger.debug(f"last_byte = {last_byte}")
        logger.debug(f"lines found = {len(lines)}")
        return last_byte, lines


def lines_to_SampleCreate(
    headers: list, lines: list, trace_id: int, delimiter: Optional[str] = None
) -> List[dict]:
    samples = []
    for line in lines:
        data = {}
        line = line.strip()  # strip \n
        for header, value in zip(headers, line.split(delimiter)):
            try:
                value = float(value)
                if value.is_integer():
                    value = int(value)
                elif math.isnan(value) or math.isinf(value):
                    value = None
            except ValueError:
                # value is not a number
                # treat as categorical
                logger.debug(f"Value is not a number: {value}")
            data[header] = value
        sample_in = {}
        sample_in["data"] = data
        sample_in["state"] = data["state"]
        sample_in["trace_id"] = trace_id
        samples.append(sample_in)
    return samples


def check_for_new_samples(db: Database, trace: schemas.Trace) -> None:
    last_byte, lines = read_lines(trace)
    in_samples = lines_to_SampleCreate(
        trace["headers_line"].split(trace["delimiter"]),
        lines,
        trace_id=trace["id"],
        delimiter=trace["delimiter"],
    )
    if in_samples:
        crud.sample.create_multi(db, objs_in=in_samples)
    # update the trace byte
    crud.trace.update(db, db_obj=trace, obj_in={"last_byte": last_byte})



from typing import Any, Dict, Generic, List, Optional, Type, TypeVar, Union
from beastiary.db.database import Database

from fastapi.encoders import jsonable_encoder
from pydantic import BaseModel

InDBSchemaType = TypeVar("InDBSchemaType", bound=BaseModel)
CreateSchemaType = TypeVar("CreateSchemaType", bound=BaseModel)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseModel)


class CRUDBase(Generic[InDBSchemaType, CreateSchemaType, UpdateSchemaType]):
    def __init__(self, model: Type[InDBSchemaType]):
        """
        CRUD object with default methods to Create, Read, Update, Delete (CRUD).
        **Parameters**
        * `model`: A Pydantic model (schema) class for in database representation
        * `schema`: A Pydantic model (schema) class
        """
        self.model = model

    def get(self, db: Database, id: Any) -> Optional[InDBSchemaType]:
        return db.query(self.model.__name__, id=id)

    def get_multi(
        self, db: Database, *, skip: int = 0, limit: int = 100
    ) -> List[InDBSchemaType]:
        return db.query(self.model.__name__, skip=skip, limit=limit)

    def create(self, db: Database, *, obj_in: CreateSchemaType) -> InDBSchemaType:
        db_obj = jsonable_encoder(obj_in)
        db.add(self.model.__name__, db_obj)
        return db_obj

    def update(
        self,
        db: Database,
        *,
        db_obj: dict,
        obj_in: Union[UpdateSchemaType, Dict[str, Any]],
    ) -> InDBSchemaType:
        db_obj.update(obj_in)
        return db_obj



class Database:
    def __init__(self) -> None:
        self.data = {}

    def __repr__(self) -> str:
        return f"Database({' '.join([f'{table}:{len(self.data[table])}'  for table in self.data])})"

    def create_table(self, name):
        if name in self.data:
            raise ValueError(f"Table {name} already exists!")
        self.data[name] = []

    def query(self, table_name: str, *, id: int = None, skip=0, limit=100, **kwargs):
        # the id will change if i remove any elements from the list...
        if not id is None:
            try:
                return self.data[table_name][id]
            except IndexError:
                return None
        return [
            row
            for row in self.data[table_name]
            if all(row[k] == v for k, v in kwargs.items())
        ][skip : skip + limit]

    def add(self, table_name: str, obj):
        self.data[table_name].append(obj)

    def add_all(self, table_name, objs: list):
        self.data[table_name].extend(objs)



from typing import Any, List, Optional
import os
from pathlib import Path
from fastapi import APIRouter, Depends

router = APIRouter()


from pydantic import BaseModel


class File(BaseModel):
    name: str
    path: str
    is_dir: bool


class Dir(BaseModel):
    path: str
    parent: str
    files: List[File]


@router.get("/", response_model=Dir)
def list_directory(path: Optional[str] = None) -> dict:
    """
    List the folders/files in a diretory
    """
    if not path:
        path = os.path.abspath(os.getcwd())
    files = [
        {
            "name": name,
            "path": os.path.join(path, name),
            "is_dir": os.path.isdir(os.path.join(path, name)),
        }
        for name in os.listdir(path)
        if not name.startswith(".")
    ]

    sorted_files = sorted(files, key=lambda d: d["is_dir"], reverse=True)

    directory = {
        "path": path,
        "parent": str(Path(path).parent.absolute()),
        "files": sorted_files,
    }
    return directory



from beastiary.api.core import add_trace, check_for_new_samples
from beastiary.log import logger

from typing import Any, List
from fastapi import APIRouter, Depends, HTTPException, Request, Path

from beastiary import crud, schemas

router = APIRouter()


@router.get("/", response_model=List[schemas.Trace])
def get_traces(
    request: Request,
    skip: int = 0,
    limit: int = 100,
) -> List[schemas.Trace]:
    """
    Retrieve traces.
    """
    traces = crud.trace.get_multi(db=request.app.db, skip=skip, limit=limit)
    return traces


@router.get("/{trace_id}", response_model=schemas.Trace)
def get_trace(request: Request, trace_id: int) -> dict:
    """
    Retrieve traces.
    """
    trace = crud.trace.get(db=request.app.db, id=trace_id)
    if not trace:
        raise HTTPException(404, "Trace not found!")
    return trace


@router.post("/", response_model=schemas.Trace)
def create_trace(
    request: Request,
    *,
    trace_in: schemas.TraceCreate,
) -> dict:
    """
    Create new trace.
    """
    try:
        trace = add_trace(request.app.db, trace_in)
    except FileNotFoundError as e:
        raise HTTPException(404, detail="Could not find log file!")
    except ValueError as e:
        raise HTTPException(400, detail=str(e))
    except Exception as e:
        print(e)
        raise HTTPException(500, detail=f"Could not add {trace_in.path}")
    return trace


@router.get("/{trace_id}/samples", response_model=List[schemas.Sample])
def get_samples(
    request: Request,
    trace_id: int,
    skip: int = 0,
    limit: int = 100,
) -> Any:
    """
    Retrieve samples.
    """
    trace = crud.trace.get(db=request.app.db, id=trace_id)
    if not trace:
        raise HTTPException(404, detail="Trace not found!")

    samples = crud.sample.get_multi_by_trace(
        request.app.db, trace_id=trace["id"], skip=skip, limit=limit
    )
    logger.debug(f"{limit} samples requested - {len(samples)} found")
    if len(samples) < limit:
        logger.debug(f"Checking for new samples in {trace['path']}")
        try:
            check_for_new_samples(request.app.db, trace=trace)
        except Exception as e:
            logger.error(str(e))
            raise HTTPException(500, detail=f"Could read samples in {trace['path']}")
        # get samples
        samples = crud.sample.get_multi_by_trace(
            request.app.db, trace_id=trace["id"], skip=skip, limit=limit
        )
    logger.debug(f"Returning {len(samples)} samples")
    return samples



from typing import List
from beastiary.db.database import Database

from fastapi.encoders import jsonable_encoder

from beastiary.crud.base import CRUDBase
from beastiary.schemas.sample import SampleCreate, SampleUpdate, Sample


class CRUDSample(CRUDBase[Sample, SampleCreate, SampleUpdate]):
    def create_multi(self, db: Database, *, objs_in: List[dict]) -> List[Sample]:
        db.add_all(self.model.__name__, objs_in)
        return objs_in

    def get_multi_by_trace(
        self, db: Database, *, trace_id: int, skip: int = 0, limit: int = 100
    ) -> List[Sample]:
        return db.query(self.model.__name__, skip=skip, limit=limit, trace_id=trace_id)


sample = CRUDSample(Sample)



from beastiary.log import logger
from typing import List
from beastiary.db import Database

from fastapi.encoders import jsonable_encoder

from beastiary.crud.base import CRUDBase
from beastiary.schemas.trace import TraceCreate, TraceUpdate, Trace


class CRUDTrace(CRUDBase[Trace, TraceCreate, TraceUpdate]):
    def create(  # type: ignore
        self, db: Database, *, obj_in: TraceCreate, headers_line: str, last_byte: int
    ) -> Trace:
        trace = obj_in.dict()
        trace.update(
            last_byte=last_byte,
            headers_line=headers_line,
            id=len(db.data[self.model.__name__]),
        )
        logger.debug(f"Adding Trace: {trace}")
        db.add(self.model.__name__, trace)
        return trace


trace = CRUDTrace(Trace)
